@@ -331,19 +331,19 @@ In a nutshell, the following issues are addressed:
 %        convert between lexical representations and Prolog terms.
 
 rdf(S,P,O) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf(S,P,O0),
     post_object(O,O0).
 
 rdf(S,P,O,G) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     pre_graph(G,G0),
     rdf_db:rdf(S,P,O0,G0),
     post_graph(G, G0),
     post_object(O,O0).
 
-%!  rdf_has(?S, ?P, ?O) is nondet.
-%!  rdf_has(?S, ?P, ?O, -RealP) is nondet.
+%!  rdf_has(?S, +P, ?O) is nondet.
+%!  rdf_has(?S, +P, ?O, -RealP) is nondet.
 %
 %   Similar to rdf/3 and rdf/4, but   P  matches all predicates that
 %   are defined as an rdfs:subPropertyOf of   P. This predicate also
@@ -351,12 +351,12 @@ rdf(S,P,O,G) :-
 %   `symmetric`. See rdf_set_predicate/2.
 
 rdf_has(S,P,O) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf_has(S,P,O0),
     post_object(O,O0).
 
 rdf_has(S,P,O,RealP) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf_has(S,P,O0,RealP),
     post_object(O,O0).
 
@@ -364,7 +364,7 @@ rdf_has(S,P,O,RealP) :-
 %!  rdf_update(+S, +P, +O, ++Action) is det.
 %!  rdf_update(+S, +P, +O, +G, ++Action) is det.
 %
-%   Replaces one of  the  three  fields   on  the  matching  triples
+%   Replaces one of the three  (four)   fields  on  the matching triples
 %   depending on Action:
 %
 %     * subject(Resource)
@@ -376,14 +376,14 @@ rdf_has(S,P,O,RealP) :-
 %     literal(Value).
 %     * graph(Graph)
 %     Moves the triple from its current named graph to Graph.
-%     This only works with rdf_update/4 and will throw an error when
-%     used with rdf_update/3.
+%     This only works with rdf_update/5 and throws an error when
+%     used with rdf_update/4.
 %
-%   The argument matching  the  action  must   be  ground.  If  this
-%   argument is equivalent to  the  current   value,  no  action  is
-%   performed. Otherwise, the requested action   is performed on all
-%   matching triples.  For example, all resources typed `rdfs:Class`
-%   can be changed to `owl:Class` using
+%   The argument matching Action must  be   ground.  If this argument is
+%   equivalent to the current value, no  action is performed. Otherwise,
+%   the requested action is  performed  on   all  matching  triples. For
+%   example,  all  resources  typed  `rdfs:Class`   can  be  changed  to
+%   `owl:Class` using
 %
 %     ```
 %     ?- rdf_update(_, rdf:type, rdfs:'Class',
@@ -455,12 +455,12 @@ rdf_update_(S, P, O, G1, graph(G2)) :-
 %   _breadth first_ search.
 
 rdf_reachable(S,P,O) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf_reachable(S,P,O0),
     post_object(O,O0).
 
 rdf_reachable(S,P,O,MaxD,D) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf_reachable(S,P,O0,MaxD,D),
     post_object(O,O0).
 
@@ -494,11 +494,11 @@ rdf_assert(S,P,O,G) :-
 %   instantiate any of its arguments.
 
 rdf_retractall(S,P,O) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf_retractall(S,P,O0).
 
 rdf_retractall(S,P,O,G) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     pre_graph(G,G0),
     rdf_db:rdf_retractall(S,P,O0,G0).
 
@@ -760,20 +760,9 @@ lang_condition(lang_matches(Pattern), Lang) :-
 %   Translated constraints are removed from object.
 
 literal_condition(Object, Cond) :-
-    var(Object),
-    !,
     get_attr(Object, rdf11, Cond0),
     best_literal_cond(Cond0, Cond, Rest),
     put_cond(Object, Rest).
-literal_condition(Text@_Lang, Cond) :-
-    get_attr(Text, rdf11, Cond0),
-    !,
-    best_literal_cond(Cond0, Cond, Rest),
-    put_cond(Text, Rest).
-literal_condition(Text^^_Type, Cond) :-
-    get_attr(Text, rdf11, Cond0),
-    best_literal_cond(Cond0, Cond, Rest),
-    put_cond(Text, Rest).
 
 %!  best_literal_cond(+Conditions, -Best, -Rest) is semidet.
 %
@@ -970,46 +959,60 @@ post_graph(G, G0:_) :-
 post_graph(G, G).
 
 
-pre_object(Literal, literal(Cond, Value)) :-
-    literal_condition(Literal, Cond),
-    !,
-    debug(literal_index, 'Search literal using ~p', [literal(Cond, Value)]),
-    literal_value0(Literal, Value).
-pre_object(Literal, literal(Value)) :-
-    literal_class(Literal, Value),
-    !,
-    debug(literal_index, 'Search literal using ~p', [literal(Value)]).
-pre_object(Var, _Var) :-
-    var(Var),
-    !.
+%   left for code calling this directly
+
 pre_object(Atom, URI) :-
+    pre_object(Atom, URI, _, _).
+
+%!  pre_object(+APIObject, -DBObject, +APISubject, +APIPredicate)
+
+pre_object(Atom, URI, _, _) :-
     atom(Atom),
     \+ boolean(Atom),
     !,
     URI = Atom.
-pre_object(Val@Lang, literal(lang(Lang, Val0))) :-
+pre_object(Var, Var1, Subj, Pred) :-
+    var(Var),
     !,
-    in_lang_string(Val, Val0).
-pre_object(Val^^Type, literal(Literal)) :-
-    !,
-    in_type(Type, Val, Type0, Val0),
-    (   var(Type0), var(Val0)
-    ->  true
-    ;   Literal = type(Type0, Val0)
+    (   literal_condition(Var, Cond)
+    ->  Var1 = literal(Cond, _)
+    ;   literal_class(Var, Value)
+    ->  Var1 = literal(Value)
+    ;   (   Var == Subj
+        ->  Var1 = Subj
+        ;   true
+        ),
+        (   Var == Pred
+        ->  Var1 = Pred
+        ;   true
+        )
     ).
-pre_object(Obj, Val0) :-
+pre_object(Val@Lang, Var1, _, _) :-
+    !,
+    (   literal_condition(Val, Cond)
+    ->  Var1 = literal(Cond, lang(Lang, _))
+    ;   literal_class(Val@Lang, Class)
+    ->  Var1 = literal(Class)
+    ;   in_lang_string(Val, Val0),
+        Var1 = literal(lang(Lang, Val0))
+    ).
+pre_object(Val^^Type, Var1, _, _) :-
+    !,
+    (   literal_condition(Val, Cond)
+    ->  Var1 = literal(Cond, type(Type, _))
+    ;   in_type(Type, Val, Type0, Val0),
+        (   var(Type0), var(Val0)
+        ->  Var1 = literal(_)
+        ;   Var1 = literal(type(Type0, Val0))
+        )
+    ).
+pre_object(Obj, Val0, _, _) :-
     ground(Obj),
     !,
     pre_ground_object(Obj, Val0).
-pre_object(Obj, _) :-
+pre_object(Obj, _, _, _) :-
     type_error(rdf_object, Obj).
 
-literal_value0(Var, _) :-
-    var(Var),
-    !.
-literal_value0(_ @Lang, lang(Lang, _)).
-literal_value0(_^^Type, type(Type, _)).
-
 
 %!  pre_ground_object(+Object, -RDF) is det.
 %
@@ -1098,9 +1101,6 @@ pre_ground_object(Atom, URI) :-
     atom(Atom),
     !,
     URI = Atom.
-%pre_ground_object(NS:Local, URI) :-            % still leaves S and P.
-%       atom(NS), atom(Local), !,
-%       rdf_global_id(NS:Local, URI).
 pre_ground_object(literal(Lit0), literal(Lit)) :-
     old_literal(Lit0, Lit),
     !.
@@ -1522,7 +1522,7 @@ out_date_time(Type, Prolog, Lexical) :-
 rdf_term(N) :-
     ground(N),
     !,
-    pre_object(N, N0),
+    pre_object(N, N0, _, _),
     visible_term(N0).
 rdf_term(N) :-
     gen_term(N).
@@ -1546,7 +1546,7 @@ rdf_literal(Term) :-
     pre_ground_object(Term, Object),
     (rdf_db:rdf(_,_,Object)->true).
 rdf_literal(Term) :-
-    pre_object(Term,literal(Lit0)),
+    pre_object(Term,literal(Lit0), _, _),
     rdf_db:rdf_current_literal(Lit0),
     (rdf_db:rdf(_,_,literal(Lit0))->true),
     post_object(Term, literal(Lit0)).
@@ -2031,7 +2031,7 @@ rdf_last(L, _) :-
 %!  rdf_estimate_complexity(?S, ?P, ?O, -Estimate) is det.
 
 rdf_estimate_complexity(S, P, O, Estimate) :-
-    pre_object(O,O0),
+    pre_object(O,O0,S,P),
     rdf_db:rdf_estimate_complexity(S,P,O0,Estimate).
 
 
