@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2016, VU University Amsterdam
+    Copyright (c)  2013-2020, VU University Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -158,6 +158,8 @@ test('get_dict/5') :-
 	get_dict(k1, [k1:v1], V, D, v2),
 	V == v1,
 	D = a{k1:v2}.
+test('get_dict/5') :-
+	gd5.
 test('get_dict/5') :-
 	\+ get_dict(k2, a{k1:v1}, _, _, v2).
 test('get_dict/5') :-
@@ -258,6 +260,13 @@ test(select, fail) :-
 test(select, R =@= _{z:3}) :-		% implicit conversion
 	select_dict([x(1)], [x(1),z(3)], R).
 
+gd5 :-
+	Dict = #{a:42},
+	get_dict(a, Dict, V, NewDict, NewV),
+	assertion(V == 42),
+	get_dict(a, NewDict, V2),
+	assertion(V2 == NewV).
+
 :- end_tests(dict_bips).
 
 :- begin_tests(dict_order).
