@@ -37,10 +37,18 @@
 #include "pl-incl.h"
 #include "os/pl-cstack.h"
 #include "pl-dict.h"
+#include <math.h>
 #ifdef HAVE_SYS_MMAN_H
 #define MMAP_STACK 1
 #include <sys/mman.h>
 #include <unistd.h>
+#ifndef MAP_ANONYMOUS
+#ifdef MAP_ANON
+#define MAP_ANONYMOUS MAP_ANON
+#else
+#define MAP_ANONYMOUS 0
+#endif
+#endif
 #endif
 
 #undef LD
@@ -1030,6 +1038,9 @@ put_double(Word at, double d, int flags ARG_LD)
 
   *at = consPtr(p, TAG_FLOAT|STG_GLOBAL);
 
+  if ( isnan(d) )
+    d = PL_nan();			/* SWI-Prolog canonical 1.5NaN */
+
   *p++ = m;
   memcpy(p, &d, sizeof(d));
   p += WORDS_PER_DOUBLE;
@@ -1327,17 +1338,6 @@ heap_gc_warn_proc(char *msg, GC_word arg)
 }
 #endif
 
-static void
-initHBase(void)
-{ void *p = GC_MALLOC(sizeof(void*));
-  uintptr_t base = (uintptr_t)p;
-
-  GC_FREE(p);				/* Keep leak-detection happy */
-  base &= ~0xfffff;			/* round down 1m */
-  GD->heap_base = base;			/* for pointer <-> int conversion */
-}
-
-
 void
 initAlloc(void)
 { static int done = FALSE;
@@ -1378,8 +1378,6 @@ properly on Linux. Don't bother with it.
 #if O_MALLOC_DEBUG
   malloc_debug(O_MALLOC_DEBUG);
 #endif
-
-  initHBase();
 }
 
 		 /*******************************
