@@ -493,7 +493,7 @@ put_sort_list(term_t l, list sl)
 
     n = sl->next;
 					/* see also linkVal() */
-    p[1] = (needsRef(*sl->item.term) ? makeRef(sl->item.term)
+    p[1] = (needsRef(*sl->item.term) ? makeRefG(sl->item.term)
 				     : *sl->item.term);
     p[0] = FUNCTOR_dot2;
     if ( n )
